
#     Copyright (C) 2023 The Chronon Authors.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

JOIN_FOLDER_NAME = 'joins'
GROUP_BY_FOLDER_NAME = 'group_bys'
STAGING_QUERY_FOLDER_NAME = 'staging_queries'
# TODO - make team part of thrift API?
TEAMS_FILE_PATH = 'teams.json'
